#coding: utf-8

import os
import pytest

import seafileapi
from tests.utils import randstring

SERVER = os.environ.get('SEAFILE_TEST_SERVER_ADDRESS', 'http://127.0.0.1:8000')
USER = os.environ.get('SEAFILE_TEST_USERNAME', 'test@seafiletest.com')
PASSWORD = os.environ.get('SEAFILE_TEST_PASSWORD', 'testtest')
ADMIN_USER = os.environ.get('SEAFILE_TEST_ADMIN_USERNAME', 'admin@seafiletest.com')
ADMIN_PASSWORD = os.environ.get('SEAFILE_TEST_ADMIN_PASSWORD', 'adminadmin')

@pytest.fixture(scope='session')
def client():
    return seafileapi.connect(SERVER, USER, PASSWORD)

@pytest.yield_fixture(scope='function')
def repo(client):
    repo_name = 'tmp-测试资料库-%s' % randstring()
    repo = client.repos.create_repo(repo_name)
    try:
        yield repo
    finally:
        repo.delete()



import os
import string
import random

def randstring(length=12):
    return ''.join(random.choice(string.lowercase) for i in range(length))

def datafile(filename):
    return os.path.join(os.path.dirname(__file__), 'data', filename)

def filesize(path):
    return os.stat(path).st_size



from seafileapi.client import SeafileApiClient


def connect(server: str, username: str, password: str, token=None, verify_ssl=True) -> SeafileApiClient:
    client = SeafileApiClient(server, username, password, token, verify_ssl)
    return client



from urllib.parse import urlencode
from typing import Optional

from seafileapi.utils import utf8lize
from seafileapi.files import SeafDir, SeafFile
from seafileapi.utils import raise_does_not_exist


class Repo:
    """
    A seafile library
    """
    def __init__(self, client, repo_id, repo_name,
                 encrypted, owner, perm):
        self.client = client
        self.id = repo_id
        self.name = repo_name
        self.encrypted = encrypted
        self.owner = owner
        self.perm = perm

    @classmethod
    def from_json(cls, client, repo_json) -> "Repo":
        repo_json = utf8lize(repo_json)

        repo_id = repo_json['id']
        repo_name = repo_json['name']
        encrypted = repo_json['encrypted']
        perm = repo_json['permission']
        owner = repo_json['owner']

        return cls(client, repo_id, repo_name, encrypted, owner, perm)

    def is_readonly(self) -> bool:
        return 'w' not in self.perm

    @raise_does_not_exist('The requested file does not exist')
    def get_file(self, path) -> Optional[SeafFile]:
        """Get the file object located in `path` in this repo.

        Return a :class:`SeafFile` object
        """
        assert path.startswith('/')
        url = f'/api2/repos/{self.id}/file/detail/'
        query = '?' + urlencode(dict(p=path))
        response = self.client.get(url + query)
        if response:
            try:
                file_json = response.json()
                if 'id' in file_json and 'size' in file_json:
                    return SeafFile(self, path, file_json['id'], file_json['size'])
            except Exception as e:
                print(e, flush=True)

    @raise_does_not_exist('The requested dir does not exist')
    def get_dir(self, path) -> Optional[SeafDir]:
        """Get the dir object located in `path` in this repo.

        Return a :class:`SeafDir` object
        """
        assert path.startswith('/')
        url = f'/api2/repos/{self.id}/dir/'
        query = '?' + urlencode(dict(p=path))
        response = self.client.get(url + query)
        if response:
            try:
                dir_id = response.headers['oid']
                dir_json = response.json()
                dir = SeafDir(self, path, dir_id)
                dir.load_entries(dir_json)
                return dir
            except Exception as e:
                print(e, flush=True)

    def delete(self):
        """Remove this repo. Only the repo owner can do this"""
        response = self.client.delete(f'/api2/repos/{self.id}')
        if response:
            if response.ok:
                print(f'status deleted: {self.id}')
        else:
            print(f'errors with delete {self.id}')


    def list_history(self):
        """List the history of this repo

        Returns a list of :class:`RepoRevision` object.
        """
        pass

    ## Operations only the repo owner can do:

    def update(self, name=None):
        """Update the name of this repo. Only the repo owner can do
        this.
        """
        pass

    def get_settings(self):
        """Get the settings of this repo. Returns a dict containing the following
        keys:

        `history_limit`: How many days of repo history to keep.
        """
        pass

    def restore(self, commit_id):
        pass


class RepoRevision:
    def __init__(self, client, repo, commit_id):
        self.client = client
        self.repo = repo
        self.commit_id = commit_id

    def restore(self):
        """Restore the repo to this revision"""
        self.repo.revert(self.commit_id)



#coding: UTF-8

import os
import pytest

from tests.utils import randstring, datafile, filesize

@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_create_delete_file_dir(repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a file
    testfile = parentdir.create_empty_file('测试文件-%s.txt' % randstring())
    assert testfile.size == 0

    entries = parentdir.ls(force_refresh=True)
    assert len(entries) == 1

    entry = entries[0]
    assert entry.path == testfile.path
    assert entry.id == testfile.id
    assert entry.size == testfile.size

    # create a folder
    testdir = parentdir.mkdir('测试目录-%s' % randstring())
    assert len(parentdir.ls(force_refresh=True)) == 2
    assert len(testdir.ls(force_refresh=True)) == 0

    direntry = [e for e in parentdir.ls(force_refresh=True) if e.isdir][0]
    assert direntry.path == testdir.path

    testfile.delete()
    assert len(parentdir.ls(force_refresh=True)) == 1
    testdir.delete()
    assert len(parentdir.ls(force_refresh=True)) == 0


@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_upload_file(repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    fname = 'aliedit.tar.gz'
    fpath = datafile(fname)
    with open(fpath, 'r') as fp:
        testfile = parentdir.upload(fp, fname)

    with open(fpath, 'r') as fp:
        fcontent = fp.read()

    assert testfile.size == filesize(fpath)
    assert testfile.name == fname
    assert testfile.repo.id == repo.id
    assert testfile.get_content() == fcontent, \
        'uploaded file content should be the same with the original file'
    entries = parentdir.ls(force_refresh=True)
    assert len(entries) == 1

    entry = entries[0]
    assert entry.path == testfile.path
    assert entry.id == testfile.id
    assert entry.size == testfile.size

    testfile.delete()
    assert len(parentdir.ls(force_refresh=True)) == 0

def test_upload_string_as_file_content(repo):
    # test pass as string as file content when upload file
    rootdir = repo.get_dir('/')
    fname = u'testfile-%s' % randstring()
    fcontent = 'line 1\nline 2\n\r'
    f = rootdir.upload(fcontent, fname)
    assert f.name == fname
    assert f.get_content() == fcontent

@pytest.mark.parametrize('parentpath', [
    '/',
    #'/测试目录一-%s' % randstring()
    '/qweqwe%s' % randstring()
])
def test_rename_file(repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a file
    testfile = parentdir.create_empty_file('测试文件-%s.txt' % randstring())
    assert testfile.size == 0

    assert len(parentdir.ls(force_refresh=True)) == 1

    # rename a file
    newfname = 'newfile.txt'
    testfile.rename(newfname)
    assert newfname == testfile.name

@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_rename_folder(repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a folder
    testfolder = parentdir.mkdir('测试文件夹-%s' % randstring())
    assert testfolder.size == 0

    assert len(parentdir.ls(force_refresh=True)) == 1

    # rename a file
    newfname = 'newfolder'
    testfolder.rename(newfname)
    assert newfname == testfolder.name

@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_copy_file(repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a file
    testfile = parentdir.create_empty_file('测试文件-%s.txt' % randstring())
    assert testfile.size == 0

    assert len(parentdir.ls(force_refresh=True)) == 1

    tempfolder = parentdir.mkdir('tempfolder_%s' % randstring())
    assert len(tempfolder.ls(force_refresh=True)) == 0
    testfile.copyTo(tempfolder.path)
    assert len(tempfolder.ls(force_refresh=True)) == 1
    assert os.path.basename(tempfolder.ls(force_refresh=True)[-1].path) == os.path.basename(testfile.path)

@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_copy_file_to_other_repo(client, repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a file
    testfile = parentdir.create_empty_file('测试文件-%s.txt' % randstring())
    assert testfile.size == 0

    assert len(parentdir.ls(force_refresh=True)) == 1

    temp_repo = client.repos.create_repo('temp_repo')
    try:
        root_dir = temp_repo.get_dir('/')
        temp_dir = root_dir.mkdir('temp_dir')
        assert len(temp_dir.ls(force_refresh=True)) == 0
        testfile.copyTo(temp_dir.path, temp_repo.id)
        assert len(temp_dir.ls(force_refresh=True)) == 1
        assert os.path.basename(temp_dir.ls(force_refresh=True)[0].path) == os.path.basename(testfile.path)
    finally:
        temp_repo.delete()

@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_copy_folder(repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a folder
    testfolder = parentdir.mkdir('测试文件夹-%s' % randstring())
    assert testfolder.size == 0
    tempfolder = parentdir.mkdir('temp-folder-%s' % randstring())
    assert tempfolder.size == 0

    assert len(tempfolder.ls(force_refresh=True)) == 0
    assert len(parentdir.ls(force_refresh=True)) == 2

    # copy a file
    testfolder.copyTo(tempfolder.path)
    assert len(tempfolder.ls(force_refresh=True)) == 1
    assert os.path.basename(tempfolder.ls(force_refresh=True)[0].path) == os.path.basename(testfolder.path)

@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_copy_folder_to_other_repo(client, repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a folder
    testfolder = parentdir.mkdir('测试文件夹-%s' % randstring())
    assert testfolder.size == 0

    assert len(parentdir.ls(force_refresh=True)) == 1

    temp_repo = client.repos.create_repo('temp_repo')
    try:
        root_folder = temp_repo.get_dir('/')
        tempfolder = root_folder.mkdir('tempfolder')

        assert len(tempfolder.ls(force_refresh=True)) == 0
        # copy a folder
        testfolder.copyTo(tempfolder.path, temp_repo.id)
        assert len(tempfolder.ls(force_refresh=True)) == 1
        assert os.path.basename(tempfolder.ls(force_refresh=True)[0].path) == os.path.basename(testfolder.path)
    finally:
        temp_repo.delete()

@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_move_file(repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a file
    testfile = parentdir.create_empty_file('测试文件-%s.txt' % randstring())
    assert testfile.size == 0

    assert len(parentdir.ls(force_refresh=True)) == 1

    tempfolder = parentdir.mkdir('tempfolder_%s' % randstring())
    assert len(tempfolder.ls(force_refresh=True)) == 0

    testfile.moveTo(tempfolder.path)
    assert testfile.path == os.path.join(tempfolder.path, os.path.basename(testfile.path))
    assert len(tempfolder.ls(force_refresh=True)) == 1

@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_move_file_to_other_repo(client, repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a file
    testfile = parentdir.create_empty_file('测试文件-%s.txt' % randstring())
    assert testfile.size == 0

    assert len(parentdir.ls(force_refresh=True)) == 1

    temp_repo = client.repos.create_repo('temp_repo')
    try:
        root_dir = temp_repo.get_dir('/')
        temp_dir = root_dir.mkdir('temp_dir')
        assert len(temp_dir.ls(force_refresh=True)) == 0
        testfile.moveTo(temp_dir.path, temp_repo.id)
        assert testfile.path == os.path.join(temp_dir.path, os.path.basename(testfile.path))
        assert len(temp_dir.ls(force_refresh=True)) == 1
        assert testfile.repo.id == temp_repo.id
    finally:
        temp_repo.delete()

@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_move_folder(repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a folder
    testfolder = parentdir.mkdir('测试文件夹-%s' % randstring())
    assert testfolder.size == 0
    tempfolder = parentdir.mkdir('temp-folder-%s' % randstring())
    assert tempfolder.size == 0

    assert len(parentdir.ls(force_refresh=True)) == 2

    # move a folder
    testfolder.moveTo(tempfolder.path)
    assert testfolder.path == os.path.join(tempfolder.path, os.path.basename(testfolder.path))

@pytest.mark.parametrize('parentpath', [
    '/',
    '/测试目录一-%s' % randstring()
])
def test_move_folder_to_other_repo(client, repo, parentpath):
    rootdir = repo.get_dir('/')
    assert len(rootdir.ls(force_refresh=True)) == 0

    if parentpath == '/':
        parentdir = rootdir
    else:
        parentdir = rootdir.mkdir(parentpath[1:])

    # create a folder
    testfolder = parentdir.mkdir('测试文件夹-%s' % randstring())
    assert testfolder.size == 0

    assert len(parentdir.ls(force_refresh=True)) == 1

    temp_repo = client.repos.create_repo('temp_repo')
    try:
        root_folder = temp_repo.get_dir('/')
        tempfolder = root_folder.mkdir('tempfolder')

        assert len(tempfolder.ls(force_refresh=True)) == 0
        # move a folder
        testfolder.moveTo(tempfolder.path, temp_repo.id)
        assert testfolder.path == os.path.join(tempfolder.path, os.path.basename(testfolder.path))
        assert testfolder.repo.id == temp_repo.id
        assert len(tempfolder.ls(force_refresh=True)) == 1
    finally:
        temp_repo.delete()



class ClientHttpError(Exception):
    """This exception is raised if the returned http response is not as
    expected"""
    def __init__(self, code, message):
        super(ClientHttpError, self).__init__()
        self.code = code
        self.message = message

    def __str__(self):
        return 'ClientHttpError[%s: %s]' % (self.code, self.message)


class OperationError(Exception):
    """Expcetion to raise when an opeartion is failed"""
    pass


class DoesNotExist(Exception):
    """Raised when not matching resource can be found."""
    def __init__(self, msg):
        super(DoesNotExist, self).__init__()
        self.msg = msg

    def __str__(self):
        return 'DoesNotExist: %s' % self.msg



from urllib.parse import urlencode
from typing import Optional, List

from seafileapi.repo import Repo
from seafileapi.utils import raise_does_not_exist

class Repos:

    def __init__(self, client: "SeafileApiClient"):
        self.client = client

    def create_repo(self, name: str, password: Optional[str] = None):
        data = {'name': name}
        if password:
            data['passwd'] = password

        response = self.client.post('/api2/repos/', data=data)
        if response:
            try:
                data = response.json()
                if 'repo_id' in data:
                    return self.get_repo(data['repo_id'])
            except Exception as e:
                print(e, flush=True)

    @raise_does_not_exist('The requested library does not exist')
    def get_repo(self, repo_id):
        """Get the repo which has the id `repo_id`.

        Raises :exc:`DoesNotExist` if no such repo exists.
        """
        response = self.client.get(f'/api2/repos/{repo_id}')
        if response:
            try:
                repo_json = response.json()
                return Repo.from_json(self.client, repo_json)
            except Exception as e:
                print(e, flush=True)

    def list_repos(self, type=None) -> Optional[List[Repo]]:
        query = ''
        if type:
            query = '?' + urlencode(dict(type=type))
        response = self.client.get(f'/api2/repos/{query}')
        if response:
            try:
                repos_json = response.json()
                return [Repo.from_json(self.client, j) for j in repos_json]
            except Exception as e:
                print(e, flush=True)



import io
import os
import posixpath
import re
from typing import Optional, Union, Any

from seafileapi.utils import querystr, utf8lize

ZERO_OBJ_ID = '0000000000000000000000000000000000000000'


class _SeafDirentBase:
    """Base class for :class:`SeafFile` and :class:`SeafDir`.

    It provides implementation of their common operations.
    """
    isdir = None

    def __init__(self, repo, path, object_id, size=0):
        """
        :param:`path` the full path of this entry within its repo, like
        "/documents/example.md"

        :param:`size` The size of a file. It should be zero for a dir.
        """
        self.client = repo.client
        self.repo = repo
        self.path = path
        self.id = object_id
        self.size = size

    @property
    def name(self):
        return posixpath.basename(self.path)

    def list_revisions(self):
        pass

    def delete(self):
        suffix = 'dir' if self.isdir else 'file'
        url = f'/api2/repos/{self.repo.id}/{suffix}/' + querystr(p=self.path)
        resp = self.client.delete(url)
        return resp

    def rename(self, newname) -> bool:
        """Change file/folder name to newname
        """
        suffix = 'dir' if self.isdir else 'file'
        url = f'/api2/repos/{self.repo.id}/{suffix}/' + querystr(p=self.path, reloaddir='true')
        postdata = {'operation': 'rename',
                    'newname': newname}
        response = self.client.post(url, data=postdata)
        succeeded = False
        if response:
            if response.status_code == 200:
                if self.isdir:
                    new_dirent = self.repo.get_dir(os.path.join(os.path.dirname(self.path), newname))
                else:
                    new_dirent = self.repo.get_file(os.path.join(os.path.dirname(self.path), newname))
                for key in self.__dict__.keys():
                    self.__dict__[key] = new_dirent.__dict__[key]
                succeeded = True
        return succeeded

    def _copy_move_task(self, operation, dirent_type, dst_dir, dst_repo_id=None):
        url = '/api/v2.1/copy-move-task/'
        src_repo_id = self.repo.id
        src_parent_dir = os.path.dirname(self.path)
        src_dirent_name = os.path.basename(self.path)
        dst_repo_id = dst_repo_id
        dst_parent_dir = dst_dir
        operation = operation
        dirent_type = dirent_type
        postdata = {'src_repo_id': src_repo_id,
                    'src_parent_dir': src_parent_dir,
                    'src_dirent_name': src_dirent_name,
                    'dst_repo_id': dst_repo_id,
                    'dst_parent_dir': dst_parent_dir,
                    'operation': operation,
                    'dirent_type': dirent_type}
        return self.client.post(url, data=postdata)

    def copyTo(self, dst_dir, dst_repo_id=None) -> bool:
        """Copy file/folder to other directory (also to a different repo)
        """
        if dst_repo_id is None:
            dst_repo_id = self.repo.id

        dirent_type = 'dir' if self.isdir else 'file'
        resp = self._copy_move_task('copy', dirent_type, dst_dir, dst_repo_id)
        if resp:
            return resp.status_code == 200
        return False

    def moveTo(self, dst_dir, dst_repo_id=None):
        """Move file/folder to other directory (also to a different repo)
        """
        if dst_repo_id is None:
            dst_repo_id = self.repo.id

        dirent_type = 'dir' if self.isdir else 'file'
        response = self._copy_move_task('move', dirent_type, dst_dir, dst_repo_id)

        succeeded = False
        if response:
            if response.status_code == 200:
                new_repo = self.client.repos.get_repo(dst_repo_id)
                dst_path = os.path.join(dst_dir, os.path.basename(self.path))
                if self.isdir:
                    new_dirent = new_repo.get_dir(dst_path)
                else:
                    new_dirent = new_repo.get_file(dst_path)
                if new_dirent:
                    for key in self.__dict__.keys():
                        self.__dict__[key] = new_dirent.__dict__[key]
                    succeeded = True
        return succeeded

    def get_share_link(self,
                       can_edit=False,
                       can_download=True,
                       password=None,
                       expire_days=None,
                       direct_link=True) -> Optional[str]:

        url = '/api/v2.1/share-links/'
        post_data = {
            "repo_id": self.repo.id,
            "path": self.path,
            "permissions": {
                "can_edit": can_edit,
                "can_download": can_download
            }
        }
        if password:
            post_data['password'] = password
        if expire_days:
            post_data['expire_days'] = expire_days
        response = self.client.post(url, data=post_data)
        if response:
            try:
                data = response.json()
                link = data['link']
                if direct_link:
                    link = link + '?dl=1'
                return link
            except Exception as e:
                print(e, flush=True)


class SeafDir(_SeafDirentBase):
    isdir = True

    def __init__(self, *args, **kwargs):
        super(SeafDir, self).__init__(*args, **kwargs)
        self.entries = None
        self.entries = kwargs.pop('entries', None)

    def ls(self, force_refresh=False):
        """List the entries in this dir.

        Return a list of objects of class :class:`SeafFile` or :class:`SeafDir`.
        """
        if self.entries is None or force_refresh:
            self.load_entries()

        return self.entries

    def share_to_user(self, email, permission) -> bool:
        url = f'/api2/repos/{self.repo.id}/dir/shared_items/' + querystr(p=self.path)
        putdata = {
            'share_type': 'user',
            'username': email,
            'permission': permission
        }
        response = self.client.put(url, data=putdata)
        if response:
            return response.status_code == 200
        else:
            print(f'errors with share: {email}, {permission}')
        return False

    def create_empty_file(self, name) -> Optional["SeafFile"]:
        """Create a new empty file in this dir.
        Return a :class:`SeafFile` object of the newly created file.
        """
        # TODO: file name validation
        path = posixpath.join(self.path, name)
        url = f'/api2/repos/{self.repo.id}/file/' + querystr(p=path, reloaddir='true')
        postdata = {'operation': 'create'}
        response = self.client.post(url, data=postdata)
        if response:
            try:
                self.id = response.headers['oid']
                self.load_entries(response.json())
                return SeafFile(self.repo, path, ZERO_OBJ_ID, 0)
            except Exception as e:
                print(e, flush=True)

    def mkdir(self, name) -> Optional["SeafDir"]:
        """Create a new sub folder right under this dir.

        Return a :class:`SeafDir` object of the newly created sub folder.
        """
        path = posixpath.join(self.path, name)
        url = f'/api2/repos/{self.repo.id}/dir/' + querystr(p=path, reloaddir='true')
        postdata = {'operation': 'mkdir'}
        response = self.client.post(url, data=postdata)
        if response:
            try:
                self.id = response.headers['oid']
                self.load_entries(response.json())
                return SeafDir(self.repo, path, ZERO_OBJ_ID)
            except Exception as e:
                print(e, flush=True)

    def upload(self, fileobj, filename: str, replace=False) -> Optional["SeafFile"]:
        """Upload a file to this folder.

        :param:fileobj :class:`File` like object
        :param:filename The name of the file

        Return a :class:`SeafFile` object of the newly uploaded file.
        """
        if isinstance(fileobj, str):
            fileobj = io.BytesIO(fileobj.encode('utf-8'))  # so so solutions
        upload_url = self._get_upload_link()
        if upload_url:
            files = {
                'file': (filename, fileobj),
                'parent_dir': self.path,
                'replace': 1 if replace else 0,
            }
            response = self.client.post(upload_url, files=files)
            if response:
                try:
                    if response.ok:
                        return self.repo.get_file(posixpath.join(self.path, filename))
                except Exception as e:
                    print(e)
            else:
                print(f'error with response upload: {filename}')
        else:
            print(f'error with upload_url for: {filename}')

    def upload_local_file(self, filepath, name=None, replace=False) -> Optional["SeafFile"]:
        """Upload a file to this folder.

        :param:filepath The path to the local file
        :param:name The name of this new file. If None, the name of the local file would be used.

        Return a :class:`SeafFile` object of the newly uploaded file.
        """
        try:
            name = name or os.path.basename(filepath)
            fp = open(filepath, 'rb')
            return self.upload(fp, name, replace)
        except Exception as e:
            print(e, flush=True)

    def _get_upload_link(self):
        url = f'/api2/repos/{self.repo.id}/upload-link/' + querystr(p=self.path)
        response = self.client.get(url)
        if response:
            try:
                data = response.text
                return re.match(r'"(.*)"', data).group(1)
            except Exception as e:
                print(e, flush=True)

    def get_uploadable_sharelink(self):
        """Generate a uploadable shared link to this dir.

        Return the url of this link.
        """
        pass

    def load_entries(self, dirents_json=None) -> None:
        if dirents_json is None:
            url = '/api2/repos/%s/dir/' % self.repo.id + querystr(p=self.path)
            dirents_json = self.client.get(url).json()

        self.entries = [self._load_dirent(entry_json) for entry_json in dirents_json]

    def _load_dirent(self, dirent_json) -> Union["SeafFile", "SeafDir"]:
        dirent_json = utf8lize(dirent_json)
        path = posixpath.join(self.path, dirent_json['name'])
        if dirent_json['type'] == 'file':
            return SeafFile(self.repo, path, dirent_json['id'], dirent_json['size'])
        else:
            return SeafDir(self.repo, path, dirent_json['id'], 0)

    @property
    def num_entries(self) -> int:
        if self.entries is None:
            self.load_entries()
        return len(self.entries) if self.entries is not None else 0

    def __str__(self):
        return 'SeafDir[repo=%s,path=%s,entries=%s]' % \
            (self.repo.id[:6], self.path, self.num_entries)

    __repr__ = __str__


class SeafFile(_SeafDirentBase):
    isdir = False

    def _get_delete_link(self) -> str:
        url: str = '/api2/repos/%s/file/' % self.repo.id + querystr(p=self.path)
        return url

    def delete_file(self) -> Optional[bytes]:
        url: str = self._get_delete_link()
        if response := self.client.delete(url):
            if response.ok:
               return response.content

    def update(self, fileobj) -> None:
        """Update the content of this file"""
        pass

    def __str__(self):
        return 'SeafFile[repo=%s,path=%s,size=%s]' % \
            (self.repo.id[:6], self.path, self.size)

    def _get_download_link(self):
        url = '/api2/repos/%s/file/' % self.repo.id + querystr(p=self.path)
        resp = self.client.get(url)
        return re.match(r'"(.*)"', resp.text).group(1)

    def get_content(self) -> bytes:
        """Get the content of the file"""
        url = self._get_download_link()
        return self.client.get(url).content

    __repr__ = __str__



import requests
import re
from typing import Optional, Callable
from sys import exit

from seafileapi.utils import urljoin, is_ascii
from seafileapi.exceptions import ClientHttpError
from seafileapi.repos import Repos


request_filename_pattern = re.compile(b'filename\*=.*')

seahub_api_auth_token = 40


class SeafileApiClient:
    """Wraps seafile web api"""
    def __init__(self, server: str,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 token: Optional[str] = None,
                 verify_ssl: bool = True):
        """Wraps various basic operations to interact with seahub http api.
        """
        self.server = server
        self.username = username
        self.password = password
        self._token = token
        self.verify_ssl = verify_ssl
        self.default_timeout = 120
        self.repos = Repos(self)
        self.groups = Groups(self)

        if token is None:
            self._get_token()

    def _get_token(self):
        data = {
            'username': self.username,
            'password': self.password,
        }
        url = urljoin(self.server, '/api2/auth-token/')
        try:
            with requests.post(url, data=data, verify=self.verify_ssl, timeout=self.default_timeout) as resp:
                if resp.status_code != 200:
                    raise ClientHttpError(resp.status_code, resp.content)
                try:
                    _token = resp.json()
                except Exception as e:
                    print(e, flush=True)
                else:
                    self._token = _token.get('token', '')
                    if len(self._token) != seahub_api_auth_token:
                        exit('The length of seahub api auth token should be 40')
        except Exception as e:
            exit(e)

    def __str__(self):
        return 'SeafileApiClient[server=%s, user=%s]' % (self.server, self.username)

    __repr__ = __str__

    def get(self, *args, **kwargs):
        return self._send_request('GET', *args, **kwargs)

    def post(self, *args, **kwargs):
        return self._send_request('POST', *args, **kwargs)

    def put(self, *args, **kwargs):
        return self._send_request('PUT', *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._send_request('DELETE', *args, **kwargs)

    def _rewrite_request(self, *args, **kwargs):
        def func(prepared_request):
            if 'files' in kwargs:
                file = kwargs['files'].get('file', None)
                if file and isinstance(file[0], str):
                    filename = file[0]
                    if not is_ascii(filename):
                        filename = (filename + '"\r').encode('utf-8')
                        prepared_request.body = request_filename_pattern.sub(b'filename="' + filename, prepared_request.body, count=1)

            return prepared_request
        return func

    def _send_request(self, method: str, url: str, *args, **kwargs) -> Optional[requests.Response]:
        if not url.startswith('http'):
            url = urljoin(self.server, url)

        headers = kwargs.get('headers', {})
        headers.setdefault('Authorization', 'Token ' + self._token)
        kwargs['headers'] = headers

        expected = kwargs.pop('expected', 200)
        if not hasattr(expected, '__iter__'):
            expected = (expected, )

        # my some dirty hack to rewrite post body
        kwargs['auth']: Optional[Callable] = self._rewrite_request(*args, **kwargs)  # hack to rewrite post body

        kwargs['method'] = method
        kwargs['url'] = url
        kwargs['verify'] = self.verify_ssl
        kwargs['timeout'] = self.default_timeout
        try:
            resp = requests.request(*args, **kwargs)
        except Exception as e:
            print(e, flush=True)
        else:
            if resp.status_code not in expected:
                msg = 'Expected %s, but get %s' % \
                      (' or '.join(map(str, expected)), resp.status_code)
                raise ClientHttpError(resp.status_code, msg)
            return resp


class Groups:
    def __init__(self, client):
        pass

    def create_group(self, name):
        pass





class Group:
    def __init__(self, client, group_id, group_name):
        self.client = client
        self.group_id = group_id
        self.group_name = group_name

    def list_memebers(self):
        pass

    def delete(self):
        pass

    def add_member(self, username):
        pass

    def remove_member(self, username):
        pass

    def list_group_repos(self):
        pass



#coding: UTF-8

import pytest

from seafileapi.exceptions import DoesNotExist
from tests.utils import randstring

def test_create_delete_repo(client):
    repo = _create_repo(client)
    repo.delete()

    with pytest.raises(DoesNotExist):
        client.repos.get_repo(repo.id)

def test_create_encrypted_repo(client):
    repo = _create_repo(client, password=randstring())
    repo.delete()
    with pytest.raises(DoesNotExist):
        client.repos.get_repo(repo.id)

def test_list_repos(client):
    repos = client.repos.list_repos()
    for repo in repos:
        assert len(repo.id) == 36

def _create_repo(client, password=None):
    repo_name = '测试资料库-%s' % randstring()
    repo = client.repos.create_repo(repo_name, password=password)

    assert repo.name == repo_name
    assert len(repo.id) == 36
    assert repo.encrypted == (password is not None)
    assert repo.owner == 'self'

    return repo



from functools import wraps
import random
import string
from typing import Any, List, Dict, Union, Callable
from urllib.parse import urlencode

from seafileapi.exceptions import ClientHttpError, DoesNotExist


def randstring(length: int) -> str:
    if length == 0:
        return ''.join(random.choice(string.ascii_lowercase) for _ in range(random.randint(1, 30)))
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(length))


def urljoin(base: str, *args) -> str:
    url = base
    if url[-1] != '/':
        url += '/'
    for arg in args:
        arg = arg.strip('/')
        url += arg + '/'
    if '?' in url:
        url = url[:-1]
    return url


def raise_does_not_exist(msg) -> Callable:
    """Decorator to turn a function that get a http 404 response to a
    :exc:`DoesNotExist` exception."""
    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ClientHttpError as e:
                if e.code == 404:
                    raise DoesNotExist(msg)
                else:
                    raise
        return wrapped
    return decorator


def to_utf8(obj) -> Any:
    return obj


def querystr(**kwargs) -> str:
    return '?' + urlencode(kwargs)


def utf8lize(obj: Any) -> Union[List, Dict, Any]:
    if isinstance(obj, dict):
        return {k: to_utf8(v) for k, v in obj.items()}

    if isinstance(obj, list):
        return [to_utf8(x) for x in obj]

    return obj


def is_ascii(text) -> bool:
    try:
        text.encode('ascii')
    except UnicodeEncodeError:
        return False
    return True



from setuptools import setup, find_packages

__version__ = '0.0.3'


setup(name='python-seafile-2022',
      version=__version__,
      license='BSD',
      description='Client interface for Seafile Web API(2022)',
      author='Igor Rumyantsev, Andrey Skhomenko',
      author_email='andrew.foma@gmail.com',
      url='https://github.com/JohnEskimSmith/python-seafile-2022',
      platforms=['Any'],
      packages=find_packages(),
      install_requires=['requests'],
      classifiers=['Development Status :: 4 - Beta',
                   'License :: OSI Approved :: BSD License',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python'],
      )
